@@ -3,16 +3,18 @@ Serialization templates for structures used in LLUDP and HTTP bodies.
 """
 
 import abc
+import collections
 import dataclasses
 import enum
 import importlib
 import logging
+import math
 import zlib
 from typing import *
 
 import hippolyzer.lib.base.serialization as se
 from hippolyzer.lib.base import llsd
-from hippolyzer.lib.base.datatypes import UUID, IntEnum, IntFlag
+from hippolyzer.lib.base.datatypes import UUID, IntEnum, IntFlag, Vector3
 from hippolyzer.lib.base.namevalue import NameValuesSerializer
 
 try:
@@ -144,6 +146,50 @@ class InventoryType(IntEnum):
         }.get(lower, lower)
 
 
+class FolderType(IntEnum):
+    TEXTURE = 0
+    SOUND = 1
+    CALLINGCARD = 2
+    LANDMARK = 3
+    CLOTHING = 5
+    OBJECT = 6
+    NOTECARD = 7
+    # We'd really like to change this to 9 since AT_CATEGORY is 8,
+    # but "My Inventory" has been type 8 for a long time.
+    ROOT_INVENTORY = 8
+    LSL_TEXT = 10
+    BODYPART = 13
+    TRASH = 14
+    SNAPSHOT_CATEGORY = 15
+    LOST_AND_FOUND = 16
+    ANIMATION = 20
+    GESTURE = 21
+    FAVORITE = 23
+    ENSEMBLE_START = 26
+    ENSEMBLE_END = 45
+    # This range is reserved for special clothing folder types.
+    CURRENT_OUTFIT = 46
+    OUTFIT = 47
+    MY_OUTFITS = 48
+    MESH = 49
+    # "received items" for MP
+    INBOX = 50
+    OUTBOX = 51
+    BASIC_ROOT = 52
+    MARKETPLACE_LISTINGS = 53
+    MARKETPLACE_STOCK = 54
+    # Note: We actually *never* create folders with that type. This is used for icon override only.
+    MARKETPLACE_VERSION = 55
+    SETTINGS = 56
+    # Firestorm folders, may not actually exist
+    FIRESTORM = 57
+    PHOENIX = 58
+    RLV = 59
+    # Opensim folders
+    MY_SUITCASE = 100
+    NONE = -1
+
+
 @se.enum_field_serializer("AgentIsNowWearing", "WearableData", "WearableType")
 @se.enum_field_serializer("AgentWearablesUpdate", "WearableData", "WearableType")
 @se.enum_field_serializer("CreateInventoryItem", "InventoryBlock", "WearableType")
@@ -177,6 +223,7 @@ def _register_permissions_flags(message_name, block_name):
 
 @se.flag_field_serializer("ObjectPermissions", "ObjectData", "Mask")
 @_register_permissions_flags("ObjectProperties", "ObjectData")
+@_register_permissions_flags("ObjectPropertiesFamily", "ObjectData")
 @_register_permissions_flags("UpdateCreateInventoryItem", "InventoryData")
 @_register_permissions_flags("UpdateTaskInventory", "InventoryData")
 @_register_permissions_flags("CreateInventoryItem", "InventoryBlock")
@@ -201,11 +248,74 @@ class Permissions(IntFlag):
     RESERVED = 1 << 31
 
 
+@se.enum_field_serializer("ObjectSaleInfo", "ObjectData", "SaleType")
+@se.enum_field_serializer("ObjectProperties", "ObjectData", "SaleType")
+@se.enum_field_serializer("ObjectPropertiesFamily", "ObjectData", "SaleType")
+@se.enum_field_serializer("ObjectBuy", "ObjectData", "SaleType")
+@se.enum_field_serializer("RezScript", "InventoryBlock", "SaleType")
+@se.enum_field_serializer("RezObject", "InventoryData", "SaleType")
+@se.enum_field_serializer("UpdateTaskInventory", "InventoryData", "SaleType")
+@se.enum_field_serializer("UpdateCreateInventoryItem", "InventoryData", "SaleType")
+class SaleInfo(IntEnum):
+    NOT = 0
+    ORIGINAL = 1
+    COPY = 2
+    CONTENTS = 3
+
+
+@se.flag_field_serializer("ParcelInfoReply", "Data", "Flags")
+class ParcelInfoFlags(IntFlag):
+    MATURE = 1 << 0
+    # You should never see adult without mature
+    ADULT = 1 << 1
+    GROUP_OWNED = 1 << 2
+
+
+@se.flag_field_serializer("MapItemRequest", "AgentData", "Flags")
+@se.flag_field_serializer("MapNameRequest", "AgentData", "Flags")
+@se.flag_field_serializer("MapBlockRequest", "AgentData", "Flags")
+@se.flag_field_serializer("MapItemReply", "AgentData", "Flags")
+@se.flag_field_serializer("MapNameReply", "AgentData", "Flags")
+@se.flag_field_serializer("MapBlockReply", "AgentData", "Flags")
+class MapImageFlags(IntFlag):
+    # No clue, honestly. I guess there's potentially different image types you could request.
+    LAYER = 1 << 1
+
+
+@se.enum_field_serializer("MapBlockReply", "Data", "Access")
+@se.enum_field_serializer("RegionInfo", "RegionInfo", "SimAccess")
+class SimAccess(IntEnum):
+    # Treated as 'unknown', usually ends up being SIM_ACCESS_PG
+    MIN = 0
+    PG = 13
+    MATURE = 21
+    ADULT = 42
+    DOWN = 254
+
+
+@se.enum_field_serializer("MapItemRequest", "RequestData", "ItemType")
+@se.enum_field_serializer("MapItemReply", "RequestData", "ItemType")
+class MapItemType(IntEnum):
+    TELEHUB = 0x01
+    PG_EVENT = 0x02
+    MATURE_EVENT = 0x03
+    # No longer supported, 2009-03-02 KLW
+    DEPRECATED_POPULAR = 0x04
+    DEPRECATED_AGENT_COUNT = 0x05
+    AGENT_LOCATIONS = 0x06
+    LAND_FOR_SALE = 0x07
+    CLASSIFIED = 0x08
+    ADULT_EVENT = 0x09
+    LAND_FOR_SALE_ADULT = 0x0a
+
+
 @se.flag_field_serializer("RezObject", "RezData", "ItemFlags")
 @se.flag_field_serializer("RezMultipleAttachmentsFromInv", "ObjectData", "ItemFlags")
 @se.flag_field_serializer("RezObject", "InventoryData", "Flags")
+@se.flag_field_serializer("RezScript", "InventoryBlock", "Flags")
 @se.flag_field_serializer("UpdateCreateInventoryItem", "InventoryData", "Flags")
 @se.flag_field_serializer("UpdateTaskInventory", "InventoryData", "Flags")
+@se.flag_field_serializer("ChangeInventoryItemFlags", "InventoryData", "Flags")
 class InventoryItemFlags(IntFlag):
     # The asset has only one reference in the system. If the
     # inventory item is deleted, or the assetid updated, then we
@@ -232,7 +342,8 @@ class InventoryItemFlags(IntFlag):
     OBJECT_HAS_MULTIPLE_ITEMS = 0x200000
 
     @property
-    def attachment_point(self):
+    def subtype(self):
+        """Subtype of the given item type, could be an attachment point or setting type, etc."""
         return self & 0xFF
 
 
@@ -761,6 +872,7 @@ class MCode(IntEnum):
 @se.flag_field_serializer("ObjectUpdateCompressed", "ObjectData", "UpdateFlags")
 @se.flag_field_serializer("ObjectUpdateCached", "ObjectData", "UpdateFlags")
 @se.flag_field_serializer("ObjectAdd", "ObjectData", "AddFlags")
+@se.flag_field_serializer("ObjectDuplicate", "SharedData", "DuplicateFlags")
 class ObjectUpdateFlags(IntFlag):
     USE_PHYSICS = 1 << 0
     CREATE_SELECTED = 1 << 1
@@ -796,6 +908,9 @@ class ObjectUpdateFlags(IntFlag):
     ZLIB_COMPRESSED_REPRECATED = 1 << 31
 
 
+JUST_CREATED_FLAGS = (ObjectUpdateFlags.CREATE_SELECTED | ObjectUpdateFlags.OBJECT_YOU_OWNER)
+
+
 class AttachmentStateAdapter(se.Adapter):
     # Encoded attachment point ID for attached objects
     # nibbles are swapped around because old attachment nums only used to live
@@ -840,6 +955,15 @@ class ObjectStateSerializer(se.AdapterSubfieldSerializer):
     ORIG_INLINE = True
 
 
+@se.subfield_serializer("ObjectUpdate", "RegionData", "TimeDilation")
+@se.subfield_serializer("ObjectUpdateCompressed", "RegionData", "TimeDilation")
+@se.subfield_serializer("ObjectUpdateCached", "RegionData", "TimeDilation")
+@se.subfield_serializer("ImprovedTerseObjectUpdate", "RegionData", "TimeDilation")
+class TimeDilationSerializer(se.AdapterSubfieldSerializer):
+    ADAPTER = se.QuantizedFloat(se.U16, 0.0, 1.0, False)
+    ORIG_INLINE = True
+
+
 @se.subfield_serializer("ImprovedTerseObjectUpdate", "ObjectData", "Data")
 class ImprovedTerseObjectUpdateDataSerializer(se.SimpleSubfieldSerializer):
     TEMPLATE = se.Template({
@@ -862,12 +986,12 @@ class ShineLevel(IntEnum):
     HIGH = 3
 
 
-@dataclasses.dataclass
+@dataclasses.dataclass(unsafe_hash=True)
 class BasicMaterials:
     # Meaning is technically implementation-dependent, these are in LL data files
-    Bump: int = se.bitfield_field(bits=5)
-    FullBright: bool = se.bitfield_field(bits=1, adapter=se.BoolAdapter())
-    Shiny: int = se.bitfield_field(bits=2, adapter=se.IntEnum(ShineLevel))
+    Bump: int = se.bitfield_field(bits=5, default=0)
+    FullBright: bool = se.bitfield_field(bits=1, adapter=se.BoolAdapter(), default=False)
+    Shiny: int = se.bitfield_field(bits=2, adapter=se.IntEnum(ShineLevel), default=0)
 
 
 BUMP_SHINY_FULLBRIGHT = se.BitfieldDataclass(BasicMaterials, se.U8)
@@ -881,12 +1005,12 @@ class TexGen(IntEnum):
     CYLINDRICAL = 0x6
 
 
-@dataclasses.dataclass
+@dataclasses.dataclass(unsafe_hash=True)
 class MediaFlags:
-    WebPage: bool = se.bitfield_field(bits=1, adapter=se.BoolAdapter())
-    TexGen: "TexGen" = se.bitfield_field(bits=2, adapter=se.IntEnum(TexGen))
+    WebPage: bool = se.bitfield_field(bits=1, adapter=se.BoolAdapter(), default=False)
+    TexGen: "TexGen" = se.bitfield_field(bits=2, adapter=se.IntEnum(TexGen), default=TexGen.DEFAULT)
     # Probably unused but show it just in case
-    _Unused: int = se.bitfield_field(bits=5)
+    _Unused: int = se.bitfield_field(bits=5, default=0)
 
 
 # Not shifted so enum definitions can match indra
@@ -1039,9 +1163,64 @@ def _te_field(spec: se.SERIALIZABLE_TYPE, first=False, optional=False,
 _T = TypeVar("_T")
 _TE_FIELD_KEY = Optional[Sequence[int]]
 
+# If this seems weird it's because it is. TE offsets are S16s with `0` as the actual 0
+# point, and LL divides by `0x7FFF` to convert back to float. Negative S16s can
+# actually go to -0x8000 due to two's complement, creating a larger range for negatives.
+TE_S16_COORD = se.QuantizedFloat(se.S16, -1.000030518509476, 1.0, False)
+
+
+class PackedTERotation(se.QuantizedFloat):
+    """Another weird one, packed TE rotations have their own special quantization"""
+
+    def __init__(self):
+        super().__init__(se.S16, math.pi * -2, math.pi * 2, zero_median=False)
+        self.step_mag = 1.0 / (se.U16.max_val + 1)
+
+    def _float_to_quantized(self, val: float, lower: float, upper: float):
+        val = math.fmod(val, upper)
+        val = super()._float_to_quantized(val, lower, upper)
+        if val == se.S16.max_val + 1:
+            val = self.prim_min
+        return val
+
 
 @dataclasses.dataclass
 class TextureEntry:
+    """Representation of a TE for a single face. Not sent over the wire."""
+    Textures: UUID = UUID('89556747-24cb-43ed-920b-47caed15465f')
+    Color: bytes = b"\xff\xff\xff\xff"
+    ScalesS: float = 1.0
+    ScalesT: float = 1.0
+    OffsetsS: float = 0.0
+    OffsetsT: float = 0.0
+    # In radians
+    Rotation: float = 0.0
+    MediaFlags: "MediaFlags" = dataclasses.field(default_factory=MediaFlags)
+    BasicMaterials: "BasicMaterials" = dataclasses.field(default_factory=BasicMaterials)
+    Glow: float = 0.0
+    Materials: UUID = UUID.ZERO
+
+    def st_to_uv(self, st_coord: Vector3) -> Vector3:
+        """Convert OpenGL ST coordinates to UV coordinates, accounting for mapping"""
+        uv = Vector3(st_coord.X - 0.5, st_coord.Y - 0.5)
+        cos_rot = math.cos(self.Rotation)
+        sin_rot = math.sin(self.Rotation)
+        uv = Vector3(
+            X=uv.X * cos_rot + uv.Y * sin_rot,
+            Y=-uv.X * sin_rot + uv.Y * cos_rot
+        )
+        uv *= Vector3(self.ScalesS, self.ScalesT)
+        return uv + Vector3(self.OffsetsS + 0.5, self.OffsetsT + 0.5)
+
+
+# Max number of TEs possible according to llprimitive (but not really true!)
+# Useful if you don't know how many faces / TEs an object really has because it's mesh
+# or something.
+MAX_TES = 45
+
+
+@dataclasses.dataclass
+class TextureEntryCollection:
     Textures: Dict[_TE_FIELD_KEY, UUID] = _te_field(
         # Plywood texture
         se.UUID, first=True, default=UUID('89556747-24cb-43ed-920b-47caed15465f'))
@@ -1049,21 +1228,65 @@ class TextureEntry:
     Color: Dict[_TE_FIELD_KEY, bytes] = _te_field(Color4(invert_bytes=True), default=b"\xff\xff\xff\xff")
     ScalesS: Dict[_TE_FIELD_KEY, float] = _te_field(se.F32, default=1.0)
     ScalesT: Dict[_TE_FIELD_KEY, float] = _te_field(se.F32, default=1.0)
-    OffsetsS: Dict[_TE_FIELD_KEY, int] = _te_field(se.S16, default=0)
-    OffsetsT: Dict[_TE_FIELD_KEY, int] = _te_field(se.S16, default=0)
-    Rotation: Dict[_TE_FIELD_KEY, int] = _te_field(se.S16, default=0)
+    OffsetsS: Dict[_TE_FIELD_KEY, float] = _te_field(TE_S16_COORD, default=0.0)
+    OffsetsT: Dict[_TE_FIELD_KEY, float] = _te_field(TE_S16_COORD, default=0.0)
+    Rotation: Dict[_TE_FIELD_KEY, float] = _te_field(PackedTERotation(), default=0.0)
     BasicMaterials: Dict[_TE_FIELD_KEY, "BasicMaterials"] = _te_field(
-        BUMP_SHINY_FULLBRIGHT, default_factory=lambda: BasicMaterials(Bump=0, FullBright=False, Shiny=0),
+        BUMP_SHINY_FULLBRIGHT, default_factory=BasicMaterials,
     )
-    MediaFlags: Dict[_TE_FIELD_KEY, "MediaFlags"] = _te_field(
-        MEDIA_FLAGS,
-        default_factory=lambda: MediaFlags(WebPage=False, TexGen=TexGen.DEFAULT, _Unused=0),
-    )
-    Glow: Dict[_TE_FIELD_KEY, int] = _te_field(se.U8, default=0)
-    Materials: Dict[_TE_FIELD_KEY, UUID] = _te_field(se.UUID, optional=True, default=UUID())
+    MediaFlags: Dict[_TE_FIELD_KEY, "MediaFlags"] = _te_field(MEDIA_FLAGS, default_factory=MediaFlags)
+    Glow: Dict[_TE_FIELD_KEY, float] = _te_field(se.QuantizedFloat(se.U8, 0.0, 1.0), default=0.0)
+    Materials: Dict[_TE_FIELD_KEY, UUID] = _te_field(se.UUID, optional=True, default=UUID.ZERO)
+
+    def unwrap(self):
+        """Return `self` regardless of whether this is lazy wrapped object or not"""
+        return self
+
+    def realize(self, num_faces: int = MAX_TES) -> List[TextureEntry]:
+        """
+        Turn the "default" vs "exception cases" wire format TE representation to per-face lookups
+        Makes it easier to get all TE details associated with a specific face
+        """
+        as_dicts = [dict() for _ in range(num_faces)]
+        for field in dataclasses.fields(self):
+            key = field.name
+            vals = getattr(self, key)
+            # Fill give all faces the default value for this key
+            for te in as_dicts:
+                te[key] = vals[None]
+            # Walk over the exception cases and replace the default value
+            for face_nums, val in vals.items():
+                # Default case already handled
+                if face_nums is None:
+                    continue
+                for face_num in face_nums:
+                    if face_num >= num_faces:
+                        raise ValueError(f"Bad value for num_faces? {face_num} >= {num_faces}")
+                    as_dicts[face_num][key] = val
+        return [TextureEntry(**x) for x in as_dicts]
+
+    @classmethod
+    def from_tes(cls, tes: List[TextureEntry]) -> "TextureEntryCollection":
+        instance = cls()
+        if not tes:
+            return instance
+
+        for field in dataclasses.fields(cls):
+            te_vals: Dict[Any, List[int]] = collections.defaultdict(list)
+            for i, te in enumerate(tes):
+                # Group values by what face they occur on
+                te_vals[getattr(te, field.name)].append(i)
+            # Make most common value the "default", everything else is an exception
+            sorted_vals = sorted(te_vals.items(), key=lambda x: len(x[1]), reverse=True)
+            default_val = sorted_vals.pop(0)[0]
+            te_vals = {None: default_val}
+            for val, face_nums in sorted_vals:
+                te_vals[tuple(face_nums)] = val
+            setattr(instance, field.name, te_vals)
+        return instance
 
 
-TE_SERIALIZER = se.Dataclass(TextureEntry)
+TE_SERIALIZER = se.Dataclass(TextureEntryCollection)
 
 
 @se.subfield_serializer("ObjectUpdate", "ObjectData", "TextureEntry")
@@ -1072,7 +1295,7 @@ TE_SERIALIZER = se.Dataclass(TextureEntry)
 @se.subfield_serializer("ObjectImage", "ObjectData", "TextureEntry")
 class TextureEntrySubfieldSerializer(se.SimpleSubfieldSerializer):
     EMPTY_IS_NONE = True
-    TEMPLATE = TE_SERIALIZER
+    TEMPLATE = se.TypedBytesGreedy(TE_SERIALIZER, empty_is_none=True, lazy=True)
 
 
 DATA_PACKER_TE_TEMPLATE = se.TypedByteArray(
@@ -1555,6 +1778,7 @@ class DeRezObjectDestination(IntEnum):
 @se.flag_field_serializer("SimStats", "RegionInfo", "RegionFlagsExtended")
 @se.flag_field_serializer("RegionInfo", "RegionInfo", "RegionFlags")
 @se.flag_field_serializer("RegionInfo", "RegionInfo3", "RegionFlagsExtended")
+@se.flag_field_serializer("MapBlockReply", "Data", "RegionFlags")
 class RegionFlags(IntFlag):
     ALLOW_DAMAGE = 1 << 0
     ALLOW_LANDMARK = 1 << 1
@@ -1600,6 +1824,7 @@ class RegionHandshakeReplyFlags(IntFlag):
 @se.flag_field_serializer("TeleportStart", "Info", "TeleportFlags")
 @se.flag_field_serializer("TeleportProgress", "Info", "TeleportFlags")
 @se.flag_field_serializer("TeleportFinish", "Info", "TeleportFlags")
+@se.flag_field_serializer("TeleportLocal", "Info", "TeleportFlags")
 @se.flag_field_serializer("TeleportLureRequest", "Info", "TeleportFlags")
 class TeleportFlags(IntFlag):
     SET_HOME_TO_TARGET = 1 << 0  # newbie leaving prelude (starter area)
@@ -1618,6 +1843,158 @@ class TeleportFlags(IntFlag):
     IS_FLYING = 1 << 13
     SHOW_RESET_HOME = 1 << 14
     FORCE_REDIRECT = 1 << 15
+    VIA_GLOBAL_COORDS = 1 << 16
+    WITHIN_REGION = 1 << 17
+
+
+@se.flag_field_serializer("AvatarPropertiesReply", "PropertiesData", "Flags")
+class AvatarPropertiesFlags(IntFlag):
+    ALLOW_PUBLISH = 1 << 0  # whether profile is externally visible or not
+    MATURE_PUBLISH = 1 << 1  # profile is "mature"
+    IDENTIFIED = 1 << 2  # whether avatar has provided payment info
+    TRANSACTED = 1 << 3  # whether avatar has actively used payment info
+    ONLINE = 1 << 4  # the online status of this avatar, if known.
+    AGEVERIFIED = 1 << 5  # whether avatar has been age-verified
+
+
+@se.flag_field_serializer("AvatarGroupsReply", "GroupData", "GroupPowers")
+@se.flag_field_serializer("AvatarGroupDataUpdate", "GroupData", "GroupPowers")
+@se.flag_field_serializer("AvatarDataUpdate", "AgentDataData", "GroupPowers")
+class GroupPowerFlags(IntFlag):
+    MEMBER_INVITE = 1 << 1  # Invite member
+    MEMBER_EJECT = 1 << 2  # Eject member from group
+    MEMBER_OPTIONS = 1 << 3  # Toggle "Open enrollment" and change "Signup Fee"
+    MEMBER_VISIBLE_IN_DIR = 1 << 47
+
+    # Roles
+    ROLE_CREATE = 1 << 4  # Create new roles
+    ROLE_DELETE = 1 << 5  # Delete roles
+    ROLE_PROPERTIES = 1 << 6  # Change Role Names, Titles, and Descriptions (Of roles the user is in, only, or any role in group?)
+    ROLE_ASSIGN_MEMBER_LIMITED = 1 << 7  # Assign Member to a Role that the assigner is in
+    ROLE_ASSIGN_MEMBER = 1 << 8  # Assign Member to Role
+    ROLE_REMOVE_MEMBER = 1 << 9  # Remove Member from Role
+    ROLE_CHANGE_ACTIONS = 1 << 10  # Change actions a role can perform
+
+    # Group Identity
+    GROUP_CHANGE_IDENTITY = 1 << 11  # Charter, insignia, 'Show In Group List', 'Publish on the web', 'Mature', all 'Show Member In Group Profile' checkboxes
+
+    # Parcel Management
+    LAND_DEED = 1 << 12  # Deed Land and Buy Land for Group
+    LAND_RELEASE = 1 << 13  # Release Land (to Gov. Linden)
+    LAND_SET_SALE_INFO = 1 << 14  # Set for sale info (Toggle "For Sale", Set Price, Set Target, Toggle "Sell objects with the land")
+    LAND_DIVIDE_JOIN = 1 << 15  # Divide and Join Parcels
+
+    # Parcel Identity
+    LAND_FIND_PLACES = 1 << 17  # Toggle "Show in Find Places" and Set Category.
+    LAND_CHANGE_IDENTITY = 1 << 18  # Change Parcel Identity: Parcel Name, Parcel Description, Snapshot, 'Publish on the web', and 'Mature' checkbox
+    LAND_SET_LANDING_POINT = 1 << 19  # Set Landing Point
+
+    # Parcel Settings
+    LAND_CHANGE_MEDIA = 1 << 20  # Change Media Settings
+    LAND_EDIT = 1 << 21  # Toggle Edit Land
+    LAND_OPTIONS = 1 << 22  # Toggle Set Home Point, Fly, Outside Scripts, Create/Edit Objects, Landmark, and Damage checkboxes
+
+    # Parcel Powers
+    LAND_ALLOW_EDIT_LAND = 1 << 23  # Bypass Edit Land Restriction
+    LAND_ALLOW_FLY = 1 << 24  # Bypass Fly Restriction
+    LAND_ALLOW_CREATE = 1 << 25  # Bypass Create/Edit Objects Restriction
+    LAND_ALLOW_LANDMARK = 1 << 26  # Bypass Landmark Restriction
+    LAND_ALLOW_SET_HOME = 1 << 28  # Bypass Set Home Point Restriction
+    LAND_ALLOW_HOLD_EVENT = 1 << 41  # Allowed to hold events on group-owned land
+    LAND_ALLOW_ENVIRONMENT = 1 << 46  # Allowed to change the environment
+
+    # Parcel Access
+    LAND_MANAGE_ALLOWED = 1 << 29  # Manage Allowed List
+    LAND_MANAGE_BANNED = 1 << 30  # Manage Banned List
+    LAND_MANAGE_PASSES = 1 << 31  # Change Sell Pass Settings
+    LAND_ADMIN = 1 << 32  # Eject and Freeze Users on the land
+
+    # Parcel Content
+    LAND_RETURN_GROUP_SET = 1 << 33  # Return objects on parcel that are set to group
+    LAND_RETURN_NON_GROUP = 1 << 34  # Return objects on parcel that are not set to group
+    LAND_RETURN_GROUP_OWNED = 1 << 48  # Return objects on parcel that are owned by the group
+
+    LAND_GARDENING = 1 << 35  # Parcel Gardening - plant and move linden trees
+
+    # Object Management
+    OBJECT_DEED = 1 << 36  # Deed Object
+    OBJECT_MANIPULATE = 1 << 38  # Manipulate Group Owned Objects (Move, Copy, Mod)
+    OBJECT_SET_SALE = 1 << 39  # Set Group Owned Object for Sale
+
+    # Accounting
+    ACCOUNTING_ACCOUNTABLE = 1 << 40  # Pay Group Liabilities and Receive Group Dividends
+
+    # Notices
+    NOTICES_SEND = 1 << 42  # Send Notices
+    NOTICES_RECEIVE = 1 << 43  # Receive Notices and View Notice History
+
+    # Proposals
+    # TODO: _DEPRECATED suffix as part of vote removal - DEV-24856:
+    PROPOSAL_START = 1 << 44  # Start Proposal
+    # TODO: _DEPRECATED suffix as part of vote removal - DEV-24856:
+    PROPOSAL_VOTE = 1 << 45  # Vote on Proposal
+
+    # Group chat moderation related
+    SESSION_JOIN = 1 << 16  # can join session
+    SESSION_VOICE = 1 << 27  # can hear/talk
+    SESSION_MODERATOR = 1 << 37  # can mute people's session
+
+    EXPERIENCE_ADMIN = 1 << 49  # has admin rights to any experiences owned by this group
+    EXPERIENCE_CREATOR = 1 << 50  # can sign scripts for experiences owned by this group
+
+    # Group Banning
+    GROUP_BAN_ACCESS = 1 << 51  # Allows access to ban / un-ban agents from a group.
+
+
+@se.flag_field_serializer("RequestObjectPropertiesFamily", "ObjectData", "RequestFlags")
+@se.flag_field_serializer("ObjectPropertiesFamily", "ObjectData", "RequestFlags")
+class ObjectPropertiesFamilyRequestFlags(IntFlag):
+    BUG_REPORT = 1 << 0
+    COMPLAINT_REPORT = 1 << 1
+    OBJECT_PAY = 1 << 2
+
+
+@se.enum_field_serializer("RequestImage", "RequestImage", "Type")
+class RequestImageType(IntEnum):
+    NORMAL = 0
+    AVATAR_BAKE = 1
+
+
+@se.enum_field_serializer("ImageData", "ImageID", "Codec")
+class ImageCodec(IntEnum):
+    INVALID = 0
+    RGB = 1
+    J2C = 2
+    BMP = 3
+    TGA = 4
+    JPEG = 5
+    DXT = 6
+    PNG = 7
+
+
+@se.enum_field_serializer("LayerData", "LayerID", "Type")
+class LayerDataType(IntEnum):
+    LAND_LAYER_CODE = ord('L')
+    WIND_LAYER_CODE = ord('7')
+    CLOUD_LAYER_CODE = ord('8')
+    WATER_LAYER_CODE = ord('W')
+
+    # <FS:CR> Aurora Sim
+    # Extended land layer for Aurora Sim
+    AURORA_LAND_LAYER_CODE = ord('M')
+    AURORA_WATER_LAYER_CODE = ord('X')
+    AURORA_WIND_LAYER_CODE = ord('9')
+    AURORA_CLOUD_LAYER_CODE = ord(':')
+
+
+@se.enum_field_serializer("ModifyLand", "ModifyBlock", "Action")
+class ModifyLandAction(IntEnum):
+    LEVEL = 0
+    RAISE = 1
+    LOWER = 2
+    SMOOTH = 3
+    NOISE = 4
+    REVERT = 5
 
 
 @se.http_serializer("RenderMaterials")
