@@ -21,7 +21,7 @@ from hippolyzer.lib.base.datatypes import TaggedUnion, UUID, TupleCoord
 from hippolyzer.lib.base.helpers import bytes_escape
 from hippolyzer.lib.base.message.message_formatting import HumanMessageSerializer
 from hippolyzer.lib.proxy.message_filter import MetaFieldSpecifier, compile_filter, BaseFilterNode, MessageFilterNode, \
-    EnumFieldSpecifier
+    EnumFieldSpecifier, MatchResult
 from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
 from hippolyzer.lib.proxy.caps import CapType, SerializedCapData
 
@@ -235,7 +235,7 @@ class AbstractMessageLogEntry(abc.ABC):
         obj = self.region.objects.lookup_localid(selected_local)
         return obj and obj.FullID
 
-    def _get_meta(self, name: str):
+    def _get_meta(self, name: str) -> typing.Any:
         # Slight difference in semantics. Filters are meant to return the same
         # thing no matter when they're run, so SelectedLocal and friends resolve
         # to the selected items _at the time the message was logged_. To handle
@@ -308,7 +308,9 @@ class AbstractMessageLogEntry(abc.ABC):
 
     def _val_matches(self, operator, val, expected):
         if isinstance(expected, MetaFieldSpecifier):
-            expected = self._get_meta(str(expected))
+            if len(expected) != 1:
+                raise ValueError(f"Can only support single-level Meta specifiers, not {expected!r}")
+            expected = self._get_meta(str(expected[0]))
             if not isinstance(expected, (int, float, bytes, str, type(None), tuple)):
                 if callable(expected):
                     expected = expected()
@@ -362,12 +364,18 @@ class AbstractMessageLogEntry(abc.ABC):
             if matcher.value or matcher.operator:
                 return False
             return self._packet_root_matches(matcher.selector[0])
-        if len(matcher.selector) == 2 and matcher.selector[0] == "Meta":
-            return self._val_matches(matcher.operator, self._get_meta(matcher.selector[1]), matcher.value)
+        if matcher.selector[0] == "Meta":
+            if len(matcher.selector) == 2:
+                return self._val_matches(matcher.operator, self._get_meta(matcher.selector[1]), matcher.value)
+            elif len(matcher.selector) == 3:
+                meta_dict = self._get_meta(matcher.selector[1])
+                if not meta_dict or not hasattr(meta_dict, 'get'):
+                    return False
+                return self._val_matches(matcher.operator, meta_dict.get(matcher.selector[2]), matcher.value)
         return None
 
-    def matches(self, matcher: "MessageFilterNode"):
-        return self._base_matches(matcher) or False
+    def matches(self, matcher: "MessageFilterNode", short_circuit=True) -> "MatchResult":
+        return MatchResult(self._base_matches(matcher) or False, [])
 
     @property
     def seq(self):
@@ -388,6 +396,14 @@ class AbstractMessageLogEntry(abc.ABC):
         xmlified = re.sub(rb" <key>", b"<key>", xmlified)
         return xmlified.decode("utf8", errors="replace")
 
+    @staticmethod
+    def _format_xml(content):
+        beautified = minidom.parseString(content).toprettyxml(indent="  ")
+        # kill blank lines. will break cdata sections. meh.
+        beautified = re.sub(r'\n\s*\n', '\n', beautified, flags=re.MULTILINE)
+        return re.sub(r'<([\w]+)>\s*</\1>', r'<\1></\1>',
+                      beautified, flags=re.MULTILINE)
+
 
 class HTTPMessageLogEntry(AbstractMessageLogEntry):
     __slots__ = ["flow"]
@@ -400,7 +416,7 @@ class HTTPMessageLogEntry(AbstractMessageLogEntry):
 
         super().__init__(region, session)
         # This was a request the proxy made through itself
-        self.meta["Injected"] = flow.request_injected
+        self.meta["Synthetic"] = flow.request_injected
 
     @property
     def type(self):
@@ -476,13 +492,17 @@ class HTTPMessageLogEntry(AbstractMessageLogEntry):
                 if not beautified:
                     content_type = self._guess_content_type(message)
                     if content_type.startswith("application/llsd"):
-                        beautified = self._format_llsd(llsd.parse(message.content))
+                        try:
+                            beautified = self._format_llsd(llsd.parse(message.content))
+                        except llsd.LLSDParseError:
+                            # Sometimes LL sends plain XML with a Content-Type of application/llsd+xml.
+                            # Try to detect that case and work around it
+                            if content_type == "application/llsd+xml" and message.content.startswith(b'<'):
+                                beautified = self._format_xml(message.content)
+                            else:
+                                raise
                     elif any(content_type.startswith(x) for x in ("application/xml", "text/xml")):
-                        beautified = minidom.parseString(message.content).toprettyxml(indent="  ")
-                        # kill blank lines. will break cdata sections. meh.
-                        beautified = re.sub(r'\n\s*\n', '\n', beautified, flags=re.MULTILINE)
-                        beautified = re.sub(r'<([\w]+)>\s*</\1>', r'<\1></\1>',
-                                            beautified, flags=re.MULTILINE)
+                        beautified = self._format_xml(message.content)
             except:
                 LOG.exception("Failed to beautify message")
 
@@ -541,6 +561,20 @@ class HTTPMessageLogEntry(AbstractMessageLogEntry):
             return "application/xml"
         return content_type
 
+    def _get_meta(self, name: str) -> typing.Any:
+        lower_name = name.lower()
+        if lower_name == "url":
+            return self.flow.request.url
+        elif lower_name == "reqheaders":
+            return self.flow.request.headers
+        elif lower_name == "respheaders":
+            return self.flow.response.headers
+        elif lower_name == "host":
+            return self.flow.request.host.lower()
+        elif lower_name == "status":
+            return self.flow.response.status_code
+        return super()._get_meta(name)
+
     def to_dict(self):
         val = super().to_dict()
         val['flow'] = self.flow.get_state()
@@ -613,7 +647,7 @@ class LLUDPMessageLogEntry(AbstractMessageLogEntry):
         super().__init__(region, session)
 
     _MESSAGE_META_ATTRS = {
-        "Injected", "Dropped", "Extra", "Resent", "Zerocoded", "Acks", "Reliable",
+        "Synthetic", "Dropped", "Extra", "Resent", "Zerocoded", "Acks", "Reliable",
     }
 
     def _get_meta(self, name: str):
@@ -671,20 +705,21 @@ class LLUDPMessageLogEntry(AbstractMessageLogEntry):
     def request(self, beautify=False, replacements=None):
         return HumanMessageSerializer.to_human_string(self.message, replacements, beautify)
 
-    def matches(self, matcher):
+    def matches(self, matcher, short_circuit=True) -> "MatchResult":
         base_matched = self._base_matches(matcher)
         if base_matched is not None:
-            return base_matched
+            return MatchResult(base_matched, [])
 
         if not self._packet_root_matches(matcher.selector[0]):
-            return False
+            return MatchResult(False, [])
 
         message = self.message
 
         selector_len = len(matcher.selector)
         # name, block_name, var_name(, subfield_name)?
         if selector_len not in (3, 4):
-            return False
+            return MatchResult(False, [])
+        found_field_keys = []
         for block_name in message.blocks:
             if not fnmatch.fnmatchcase(block_name, matcher.selector[1]):
                 continue
@@ -693,13 +728,13 @@ class LLUDPMessageLogEntry(AbstractMessageLogEntry):
                     if not fnmatch.fnmatchcase(var_name, matcher.selector[2]):
                         continue
                     # So we know where the match happened
-                    span_key = (message.name, block_name, block_num, var_name)
+                    field_key = (message.name, block_name, block_num, var_name)
                     if selector_len == 3:
                         # We're just matching on the var existing, not having any particular value
                         if matcher.value is None:
-                            return span_key
-                        if self._val_matches(matcher.operator, block[var_name], matcher.value):
-                            return span_key
+                            found_field_keys.append(field_key)
+                        elif self._val_matches(matcher.operator, block[var_name], matcher.value):
+                            found_field_keys.append(field_key)
                     # Need to invoke a special unpacker
                     elif selector_len == 4:
                         try:
@@ -710,15 +745,21 @@ class LLUDPMessageLogEntry(AbstractMessageLogEntry):
                         if isinstance(deserialized, TaggedUnion):
                             deserialized = deserialized.value
                         if not isinstance(deserialized, dict):
-                            return False
+                            continue
                         for key in deserialized.keys():
                             if fnmatch.fnmatchcase(str(key), matcher.selector[3]):
                                 if matcher.value is None:
-                                    return span_key
-                                if self._val_matches(matcher.operator, deserialized[key], matcher.value):
-                                    return span_key
+                                    # Short-circuiting checking individual subfields is fine since
+                                    # we only highlight fields anyway.
+                                    found_field_keys.append(field_key)
+                                    break
+                                elif self._val_matches(matcher.operator, deserialized[key], matcher.value):
+                                    found_field_keys.append(field_key)
+                                    break
 
-        return False
+                    if short_circuit and found_field_keys:
+                        return MatchResult(True, found_field_keys)
+        return MatchResult(bool(found_field_keys), found_field_keys)
 
     @property
     def summary(self):
