@@ -11,7 +11,8 @@ import multidict
 
 from hippolyzer.lib.base.datatypes import Vector3, UUID
 from hippolyzer.lib.base.helpers import proxify
-from hippolyzer.lib.base.message.message import Message
+from hippolyzer.lib.base.message.llsd_msg_serializer import LLSDMessageSerializer
+from hippolyzer.lib.base.message.message import Message, Block
 from hippolyzer.lib.base.message.message_handler import MessageHandler
 from hippolyzer.lib.base.objects import handle_to_global_pos
 from hippolyzer.lib.client.state import BaseClientRegion
@@ -110,7 +111,8 @@ class ProxiedRegion(BaseClientRegion):
         Wrap an existing, non-unique cap with a unique URL
 
         caps like ViewerAsset may be the same globally and wouldn't let us infer
-        which session / region the request was related to without a wrapper
+        which session / region the request was related to without a wrapper URL
+        that we inject into the seed response sent to the viewer.
         """
         parsed = list(urllib.parse.urlsplit(self.caps[name][1]))
         seed_id = self.caps["Seed"][1].split("/")[-1].encode("utf8")
@@ -120,22 +122,24 @@ class ProxiedRegion(BaseClientRegion):
         # to be secure. This should save on expensive TLS context setup for each req.
         parsed[0] = "http"
         wrapper_url = urllib.parse.urlunsplit(parsed)
-        self.caps.add(name + "ProxyWrapper", (CapType.WRAPPER, wrapper_url))
-        self._recalc_caps()
+        # Register it with "ProxyWrapper" appended so we don't shadow the real cap URL
+        # in our own view of the caps
+        self.register_cap(name + "ProxyWrapper", wrapper_url, CapType.WRAPPER)
         return wrapper_url
 
     def register_proxy_cap(self, name: str):
-        """
-        Register a cap to be completely handled by the proxy
-        """
+        """Register a cap to be completely handled by the proxy"""
+        if name in self.caps:
+            # If we have an existing cap then we should just use that.
+            cap_data = self.caps[name]
+            if cap_data[1] == CapType.PROXY_ONLY:
+                return cap_data[0]
         cap_url = f"http://{uuid.uuid4()!s}.caps.hippo-proxy.localhost"
-        self.caps.add(name, (CapType.PROXY_ONLY, cap_url))
-        self._recalc_caps()
+        self.register_cap(name, cap_url, CapType.PROXY_ONLY)
         return cap_url
 
-    def register_temporary_cap(self, name: str, cap_url: str):
-        """Register a Cap that only has meaning the first time it's used"""
-        self.caps.add(name, (CapType.TEMPORARY, cap_url))
+    def register_cap(self, name: str, cap_url: str, cap_type: CapType = CapType.NORMAL):
+        self.caps.add(name, (cap_type, cap_url))
         self._recalc_caps()
 
     def resolve_cap(self, url: str, consume=True) -> Optional[Tuple[str, str, CapType]]:
@@ -169,9 +173,26 @@ class EventQueueManager:
         self._region = weakref.proxy(region)
         self._last_ack: Optional[int] = None
         self._last_payload: Optional[Any] = None
+        self.llsd_message_serializer = LLSDMessageSerializer()
+
+    def inject_message(self, message: Message):
+        self.inject_event(self.llsd_message_serializer.serialize(message, True))
 
     def inject_event(self, event: dict):
         self._queued_events.append(event)
+        if self._region:
+            circuit: ProxiedCircuit = self._region.circuit
+            session: Session = self._region.session()
+            # Inject an outbound PlacesQuery message so we can trigger an inbound PlacesReply
+            # over the EQ. That will allow us to shove our own event onto the response once it comes in,
+            # otherwise we have to wait until the EQ legitimately returns 200 due to a new event.
+            # May or may not work in OpenSim.
+            circuit.send_message(Message(
+                'PlacesQuery',
+                Block('AgentData', AgentID=session.agent_id, SessionID=session.id, QueryID=UUID()),
+                Block('TransactionData', TransactionID=UUID()),
+                Block('QueryData', QueryText=b'', QueryFlags=64, Category=-1, SimName=b''),
+            ))
 
     def take_injected_events(self):
         events = self._queued_events
